/* cresample.c
 * cmix version of resample
 * Copyright (c) 1991, Julius O. Smith III.  All rights reserved. 
 *
 *      06/07/88/clf    Initial C version created from jos SAIL version.
 *	04/23/90/mtm	Made function prototypes.
 *			Made local functions static.
 *			Use #import.
 *      03/06/91 mcnabb Modified to read & write NeXT mono & stereo 
 *                      sound files
 *      06/06/91 jos	Modification to contain default filter file.
 *			Switched from prompts to command-line options.
 *			Modified filter file format to include Nmult.
 *			Added choice between "small" and "large" filters.
 *      06/18/91 jos	Split resample.c into itself plus resamplesubs.c
 *
 *	Tue Sep  5 20:38:16 EST 1995 - amr - converted it to Cmix format to
 *                                 to get rid of the NeXT sound limitations.
 */
    /*
     *      Implements sampling rate conversions by (almost) arbitrary factors.
     *      The program internally uses 16-bit data and 16-bit filter
     *      coefficients.
     *
     *      Reference: "A Flexible Sampling-Rate Conversion Method,"
     *      J. O. Smith and P. Gossett, ICASSP, San Diego, 1984, Pgs 19.4.
     *
     *      * Need overflow detection in Filter() routines.  Also, we want
     *        to saturate instead of wrap-around and report number of clipped
     *        samples.
     */
    
    /* CHANGES from original SAIL program:
     *
     * 1. LpScl is scaled by factor (when factor<1) in resample() so this is
     *       done whether the filter was loaded or created.
     * 2. makeFilter() - ImpD[] is created from Imp[] instead of ImpR[], to 
     *       avoid problems with round-off errors.
     * 3. makeFilter() - ImpD[Nwing-1] gets NEGATIVE Imp[Nwing-1].
     * 4. SrcU/D() - Switched order of making guard bits (v>>Nhg) and
     *       normalizing.  This was done to prevent overflow.
     */
    
    /* LIBRARIES needed:
     *
     * 1. filterkit
     *       readFilter() - reads standard filter file
     *       FilterUp()   - applies filter to sample when factor >= 1
     *       FilterUD()   - applies filter to sample for any factor
     *
     * 2. math
     */

#include "cresample.h"
#include "../H/ugens.h"
#include "../H/sfheader.h"
#include <stdio.h>
//#include <math.h>
#include "filterkit.h"
//#include <sound/sound.h>
#include "cresample.h"
#include "smallfilter.h"
#include "largefilter.h"
//#define IBUFFSIZE 8192		/* Input buffer size */

//--------- Prototypes ---------------------

extern int setnote(float,float,int);	/* position output file*/
void profile(void);
static void fail(char *s);
double resample(float *p,int n_args,double *pp);

static int readData(
		    int inCount,
		    HWORD *outPtr1, // X1 temp buffers
		    HWORD *outPtr2, // X2 temp buffers
		    int dataArraySize, 
		    int nChans, int Xoff);

//static int SrcUp(HWORD X[], HWORD Y[], double factor, UWORD *Time,
//		 UHWORD Nx, UHWORD Nwing, UHWORD LpScl,
//		 HWORD Imp[], HWORD ImpD[], BOOL Interp);

//static int SrcUD(HWORD X[], HWORD Y[], double factor, UWORD *Time,
//		 UHWORD Nx, UHWORD Nwing, UHWORD LpScl,
//		 HWORD Imp[], HWORD ImpD[], BOOL Interp);

static int SrcUp(HWORD X[], float Y[], double factor, UWORD *Time,
		 UHWORD Nx, UHWORD Nwing, UHWORD LpScl,
		 HWORD Impp[], HWORD ImpD[], BOOL Interp);

static int SrcUD(HWORD X[], float Y[], double factor, UWORD *Time,
		 UHWORD Nx, UHWORD Nwing, UHWORD LpScl,
		 HWORD Imp[], HWORD ImpD[], BOOL Interp);

int doresample(		/* number of output sample returned */
    double factor,		/* factor = Sndout/Sndin */
    int inCount,		/* number of input samples to convert */
    int outCount,		/* number of output samples to compute */
    int nChans,			/* number of sound channels (1 or 2) */
    BOOL interpFilt,		/* TRUE means interpolate filter coeffs */
    BOOL largeFilter,		/* TRUE means use 65-tap FIR filter */
    char *filterFile		/* NULL for internal filter, else filename */);

static int resampleWithFilter(  /* number of output sample returned */
    double factor,		/* factor = Sndout/Sndin */
    int inCount,		/* number of input samples to convert */
    int outCount,		/* number of output samples to compute */
    int nChans,			/* number of sound channels (1 or 2) */
    BOOL interpFilt,		/* TRUE means interpolate filter coeffs */
    HWORD Imp[], HWORD ImpD[],
    UHWORD LpScl, UHWORD Nmult, UHWORD Nwing);

static int err_ret(char *s);

//------------------------------------------------------------------------
extern	SFHEADER	sfdesc[NFILES];  /* to access header information */
#define IBUFFSIZE 1024             /* Input buffer size */

int NBYTES = 32768;
static int trace = 1;		/* for debugging */
char resampleVERSION[] = "2.0 (Fri Sep 1 18:24:47 EST 1995 - amr)";
int inCount, outCount, outCountReal;
int 	input = 0, output = 1;
float xin[4];
HWORD *X1, *X2;
float  *Y1, *Y2; /* I/O buffers */
int OBUFFSIZE;

#define USAGE "\
\n\
USAGE: \n\
\n\
resample(inskip,outskip,dur,factor,semitone,quality,noninterlaced,trace,filter file)\n\
\n\
"

static void fail(char *s)
{
    fprintf(stderr,"resample: %s \n\n",s); /* Display error message  */
    fprintf(stderr,USAGE);
    exit(1);			/* Exit, indicating error */
}

double
resample(float *p,int n_args, double *pp)
{
	double semitones;		/* semitones = semitones up or down */
	BOOL interpFilt;		/* TRUE means interpolate filter coeffs */
	BOOL largeFilter;		/* TRUE means use 65-tap FIR filter */
	int nChans,outCount;
	double factor;		/* factor = Sndout/Sndin */
	char filterFile[512];

	/* Set defaults */

	largeFilter = FALSE;
	factor = -2.0;
	interpFilt = TRUE;

	/*	p0=inskip, p1=outskip, p2= duration, p3=factor, p4=factor in semitones,
			p5=quality, p6=noninterlaced, p7=trace,p8=filter
	*/

	if (n_args < 9) fail("");

	if(pp[7]>1.0){				/* -trace n */
		trace = pp[7];
		if (trace) printf("trace variable set to 0x%X\n",trace);
	}
	if(pp[5]>0.0){			       /* -aaa */
	    largeFilter = TRUE;
	    if (trace) printf("Choosing higher quality filter.\n");
	}
	if(pp[3]>0.0){			       /* -by factor */
	    factor=pp[3];
	    if (trace) printf("Sampling-rate conversion factor set to %f.\n",factor);
	}
  if(pp[4]>0.0){                              /* -by semitones */
		semitones=pp[4];
		factor = pow(2.,-semitones/12.);
		if (trace)
			printf("Sampling-rate conversion factor set to %f, %g semitones. \n",factor,semitones);
		if (pp[3]>0.0)
			printf("WARNING: Overriding factor value in pp[3] %f \n",pp[3]);
		}
	if(pp[6]>0.0){			       /* -noInterpolation */
		interpFilt = FALSE;
		if (trace) printf("Filter-table interpolation disabled.\n");
	}
	if(pp[8]>0){			       /* -filter filterFile */
exit(0);
/*	    char *c = *++argv;
		if (*c == '-') {
		c = (*--argv)+1;	// assume '-ffile' with no space
		if (*c == ' ')
		  fail("");
	    }
	    strcpy(filterFile,c);
	    if (trace) printf("Filter file set to %s.\n",filterFile); */
	 }

	if (factor < 0) {
		factor = -factor;
		factor = GetDouble("Sampling-rate conversion factor",factor,"");
	}

	inCount=setnote(pp[0],pp[2],output);	/* position output file*/
	setnote(pp[1],pp[2],input);  /* position input file*/
//printf("inCount %d %f %f %f\n", inCount, pp[0],pp[1],pp[2]);
  nChans = sfchans(&sfdesc[input]);
	outCount = ((inCount)*factor)+1;
	OBUFFSIZE = IBUFFSIZE*factor+2.0;

	X1 = (HWORD *) calloc(IBUFFSIZE,sizeof(HWORD));
 	X2 = (HWORD *) calloc(IBUFFSIZE,sizeof(HWORD));
	Y1 = (float *) calloc(OBUFFSIZE,sizeof(float));
	Y2 = (float *) calloc(OBUFFSIZE,sizeof(float));

	printf("\nStarting Conversion...inCount %d, outCount %d\n",inCount, outCount);
	outCountReal = doresample(factor, inCount, outCount, nChans,
			interpFilt, largeFilter, filterFile);
	endnote(1);
	if (outCountReal <= 0) fail("Bad parameters");

	if (trace && (outCount != outCountReal))
      fprintf(stderr,
	      "outCount = %d, outCountReal = %d\n",outCount,outCountReal);
	
	printf("...Conversion Finished:  %d output samples.\n\n",outCount);

	return(0);
}

static int readData(
		    int inCount,
		    HWORD *outPtr1, 
		    HWORD *outPtr2, 
		    int dataArraySize, 
		    int nChans, int Xoff)
    /* return: 0 - notDone */
    /*        >0 - index of last sample */
{
    int Nsamps, val=0;
    HWORD *dataStart;
    dataStart = outPtr1;
    Nsamps = dataArraySize - Xoff; /* Calculate number of samples to get */
    outPtr1 += Xoff;		/* Start at designated sample number */
	if (nChans==1) {
		for (; Nsamps>0; Nsamps--) {
//	    if (!BGETIN(outPtr1,input,Nsamps)) break; // test this....
	    if (!GETIN(xin,input)) break;
	    *outPtr1++ = (int)xin[0];
		}
   } else {
		outPtr2 += Xoff;		/* Start at designated sample number */
		for (; Nsamps>0; Nsamps--) {
	    if (!GETIN(xin,input)) break;
	    *outPtr1++ = (HWORD)xin[0];
	    *outPtr2++ = (HWORD)xin[1];
	}
    }
	if (Nsamps > 0) {
		val = outPtr1 - dataStart; /* (Calc return value) */
		while (--Nsamps >= 0) {	/*   fill unread spaces with 0's */
	    *outPtr1++ = 0;	/*   and return FALSE */
	    if (nChans==2) *outPtr2++ = 0;
		}
	}
	return(val);
}

static int resampleWithFilter(  /* number of output sample returned */
    double factor,		/* factor = Sndout/Sndin */
//    HWORD *inPtr,		/* input data pointer */
//    HWORD *outPtr,		/* output data pointer */
    int inCount,		/* number of input samples to convert */
    int outCount,		/* number of output samples to compute */
    int nChans,			/* number of sound channels (1 or 2) */
    BOOL interpFilt,		/* TRUE means interpolate filter coeffs */
    HWORD Imp[], HWORD ImpD[],
    UHWORD LpScl, UHWORD Nmult, UHWORD Nwing)
{
    UWORD Time, Time2;		/* Current time/pos in input sample */
    UHWORD Xp, Ncreep, Xoff, Xread;
//#   define OBUFFSIZE ((int)(((double)IBUFFSIZE)*factor+2.0))
//    HWORD X1[IBUFFSIZE], Y1[OBUFFSIZE]; /* I/O buffers */
//    float xout[OBUFFSIZE*nChans]; /* final output buffer */
    float xout[4]; /* std cmix output */
//    HWORD X2[IBUFFSIZE], Y2[OBUFFSIZE]; /* I/O buffers */
    UHWORD Nout, Nx;
//    HWORD *inPtrRun = inPtr;	/* Running pointer thru input */
    int i, outcnt=0, Ycount, last;
    
    /* Account for increased filter gain when using factors less than 1 */
    if (factor < 1)
      LpScl = LpScl*factor + 0.5;
    /* Calc reach of LP filter wing & give some creeping room */
    Xoff = ((Nmult+1)/2.0) * MAX(1.0,1.0/factor) + 10;
//printf("Xoff %d, Nmult %d, factor %f\n",Xoff, Nmult, factor);
    if (IBUFFSIZE < 2*Xoff)      /* Check input buffer size */
      return err_ret("IBUFFSIZE (or factor) is too small");
    Nx = IBUFFSIZE - 2*Xoff;     /* # of samples to process each iteration */
    
    last = 0;			/* Have not read last input sample yet */
    Ycount = 0;			/* Current sample and length of output file */
    Xp = Xoff;			/* Current "now"-sample pointer for input */
    Xread = Xoff;		/* Position in input array to read into */
    Time = (Xoff<<Np);		/* Current-time pointer for converter */
//printf("Xoff %d, Time %d, Nx %d, Xp %d, Xread %d\n",Xoff, Time, Nx, Xp, Xread);
    for (i=0; i<Xoff; X1[i++]=0); /* Need Xoff zeros at begining of sample */
    for (i=0; i<Xoff; X2[i++]=0); /* Need Xoff zeros at begining of sample */
        
    do {
	if (!last)		/* If haven't read last sample yet */
	{
	    last = readData(inCount, X1, X2, IBUFFSIZE,nChans, (int)Xread);
	    if (last && (last-Xoff<Nx)) { /* If last sample has been read... */
		Nx = last-Xoff;	/* ...calc last sample affected by filter */
		if (Nx <= 0)
		  break;
	    }
	}
	/* Resample stuff in input buffer */
	Time2 = Time;
	if (factor >= 1) {	/* SrcUp() is faster if we can use it */
	    Nout=SrcUp(X1,Y1,factor,&Time,Nx,Nwing,LpScl,Imp,ImpD,interpFilt);
	    if (nChans==2)
	      Nout=SrcUp(X2,Y2,factor,&Time2,Nx,Nwing,LpScl,Imp,ImpD,
			 interpFilt);
	}
	else {
	    Nout=SrcUD(X1,Y1,factor,&Time,Nx,Nwing,LpScl,Imp,ImpD,interpFilt);
	    if (nChans==2)
	      Nout=SrcUD(X2,Y2,factor,&Time2,Nx,Nwing,LpScl,Imp,ImpD,
			 interpFilt);
	}

	Time -= (Nx<<Np);	/* Move converter Nx samples back in time */
	Xp += Nx;		/* Advance by number of samples processed */
	Ncreep = (Time>>Np) - Xoff; /* Calc time accumulation in Time */

	if (Ncreep) {
	    Time -= (Ncreep<<Np);    /* Remove time accumulation */
	    Xp += Ncreep;            /* and add it to read pointer */
	}
	for (i=0; i<IBUFFSIZE-Xp+Xoff; i++) { /* Copy part of input signal */
	    X1[i] = X1[i+Xp-Xoff]; /* that must be re-used */
	    if (nChans==2)
	      X2[i] = X2[i+Xp-Xoff]; /* that must be re-used */
	}
	if (last) {		/* If near end of sample... */
	    last -= Xp;		/* ...keep track were it ends */
	    if (!last)		/* Lengthen input by 1 sample if... */
	      last++;		/* ...needed to keep flag TRUE */
	}
	Xread = i;		/* Pos in input buff to read new data into */
	Xp = Xoff;
	
//				printf("outcnt %d, Ycount %d, outCount %d\n",outcnt, Ycount, outCount);
	Ycount += Nout;
	if (Ycount>outCount) {
	    Nout -= (Ycount-outCount);
	    Ycount = outCount;
	}

	if (Nout > OBUFFSIZE) /* Check to see if output buff overflowed */
	  return err_ret("Output array overflow");
	outcnt=Nout;
	if (nChans==1) {
			register float *y1p = &(Y1[0]);
//	    register HWORD *y1p = &(Y1[0]);
//				BADDOUT(Y1,1,outcnt);
	    while (Nout--){	/* Nout is UNSIGNED */
	      xout[0] = (float)*y1p++;
				ADDOUT(xout,1);}
//				BADDOUT(xout,1,outcnt);
//	    outPtr = op;
	} else {
			register float *y1p = &(Y1[0]);
			register float *y2p = &(Y2[0]);
//	    register HWORD *y1p = &(Y1[0]);
//	    register HWORD *y2p = &(Y2[0]);
	    while (Nout--) {
				xout[0] = (float)*y1p++; // have to merge chans
				xout[1] = (float)*y2p++;
				ADDOUT(xout,1);
	    }
//	    outPtr = op;
//		BADDOUT(xout,1,outcnt);
	}
    } while (Ycount<outCount); /* Continue until done */
    return(Ycount);		/* Return # of samples in output file */
}



int doresample(			/* number of output sample returned */
    double factor,		/* factor = Sndout/Sndin */
    int inCount,		/* number of input samples to convert */
    int outCount,		/* number of output samples to compute */
    int nChans,			/* number of sound channels (1 or 2) */
    BOOL interpFilt,		/* TRUE means interpolate filter coeffs */
    BOOL largeFilter,		/* TRUE means use 65-tap FIR filter */
    char *filterFile		/* NULL for internal filter, else filename */
	 )
{
    UHWORD LpScl;		/* Unity-gain scale factor */
    UHWORD Nwing;		/* Filter table size */
    UHWORD Nmult;		/* Filter length for up-conversions */
    HWORD *Imp=0;		/* Filter coefficients */
    HWORD *ImpD=0;		/* ImpD[n] = Imp[n+1]-Imp[n] */
    
    /* Check for illegal constants */
    if (Np >= 16)
      return err_ret("Error: Np>=16");
    if (Nb+Nhg+NLpScl >= 32)
      return err_ret("Error: Nb+Nhg+NLpScl>=32");
    if (Nh+Nb > 32)
      return err_ret("Error: Nh+Nb>32");
    
    /* Set defaults */

    if (filterFile != NULL && *filterFile != '\0') {
	if (readFilter(filterFile, &Imp, &ImpD, &LpScl, &Nmult, &Nwing))
	  return err_ret("could not find filter file, "
	       "or syntax error in contents of filter file");
    } else if (largeFilter) {
	Nmult = LARGE_FILTER_NMULT;
	Imp = LARGE_FILTER_IMP;	/* Impulse response */
	ImpD = LARGE_FILTER_IMPD;	/* Impulse response deltas */
	LpScl = LARGE_FILTER_SCALE;	/* Unity-gain scale factor */
	Nwing = LARGE_FILTER_NWING;	/* Filter table length */
    } else {
	Nmult = SMALL_FILTER_NMULT;
	Imp = SMALL_FILTER_IMP;	/* Impulse response */
	ImpD = SMALL_FILTER_IMPD;	/* Impulse response deltas */
	LpScl = SMALL_FILTER_SCALE;	/* Unity-gain scale factor */
	Nwing = SMALL_FILTER_NWING;	/* Filter table length */
    }
    return resampleWithFilter(factor,inCount,outCount,nChans, 
			      interpFilt, Imp, ImpD, LpScl, Nmult, Nwing);
}

/* Sampling rate up-conversion only subroutine;
 * Slightly faster than down-conversion;
 */
static int SrcUp(HWORD X[], float Y[], double factor, UWORD *Time,
		 UHWORD Nx, UHWORD Nwing, UHWORD LpScl,
		 HWORD Imp[], HWORD ImpD[], BOOL Interp)
{
    HWORD *Xp;
    float *Ystart;
    WORD v;
    double dt;                  /* Step through input signal */ 
    UWORD dtb;                  /* Fixed-point version of Dt */
    UWORD endTime;              /* When Time reaches EndTime, return to user */
    
    dt = 1.0/factor;            /* Output sampling period */
    dtb = dt*(1<<Np) + 0.5;     /* Fixed-point representation */
    
    Ystart = Y;
    endTime = *Time + (1<<Np)*(WORD)Nx;
    while (*Time < endTime)
    {
	Xp = &X[*Time>>Np];      /* Ptr to current input sample */
	/* Perform left-wing inner product */
	v = FilterUp(Imp, ImpD, Nwing, Interp, Xp, (HWORD)(*Time&Pmask),-1);
	/* Perform right-wing inner product */
	v += FilterUp(Imp, ImpD, Nwing, Interp, Xp+1, (HWORD)((-*Time)&Pmask),1);
	v >>= Nhg;               /* Make guard bits */
	v *= LpScl;              /* Normalize for unity filter gain */
	*Y++ = v>>NLpScl;        /* Deposit output */
	*Time += dtb;            /* Move to next sample by time increment */
    }
    return (Y - Ystart);        /* Return the number of output samples */
}

/* Sampling rate conversion subroutine */

static int SrcUD(HWORD X[], float Y[], double factor, UWORD *Time,
		 UHWORD Nx, UHWORD Nwing, UHWORD LpScl,
		 HWORD Imp[], HWORD ImpD[], BOOL Interp)
{
    HWORD *Xp;
    float *Ystart;
    WORD v;
    
    double dh;                  /* Step through filter impulse response */
    double dt;                  /* Step through input signal */
    UWORD endTime;              /* When Time reaches EndTime, return to user */
    UWORD dhb, dtb;             /* Fixed-point versions of Dh,Dt */
    
    dt = 1.0/factor;            /* Output sampling period */
    dtb = dt*(1<<Np) + 0.5;     /* Fixed-point representation */
    
    dh = MIN(Npc, factor*Npc);  /* Filter sampling period */
    dhb = dh*(1<<Na) + 0.5;     /* Fixed-point representation */
    
    Ystart = Y;
    endTime = *Time + (1<<Np)*(WORD)Nx;
    while (*Time < endTime)
    {
	Xp = &X[*Time>>Np];      /* Ptr to current input sample */
	v = FilterUD(Imp, ImpD, Nwing, Interp, Xp, (HWORD)(*Time&Pmask),
		     -1, dhb);            /* Perform left-wing inner product */
	v += FilterUD(Imp, ImpD, Nwing, Interp, Xp+1, (HWORD)((-*Time)&Pmask),
		      1, dhb);            /* Perform right-wing inner product */
	v >>= Nhg;               /* Make guard bits */
	v *= LpScl;              /* Normalize for unity filter gain */
	*Y++ = v>>NLpScl;        /* Deposit output */
	*Time += dtb;            /* Move to next sample by time increment */
    }
    return (Y - Ystart);        /* Return the number of output samples */
}

static int err_ret(char *s)
{
    fprintf(stderr,"resample: %s \n\n",s); /* Display error message  */
    return -1;
}

void profile(void)
{
	UG_INTRO("resample",resample)
}
